/* -*- mode:prolog; -*-
 ******************************************************************
 * $Id: tag_compile.pl 718 2008-09-09 16:19:00Z clerger $
 * Copyright (C) 1998, 2002, 2003, 2004, 2005, 2006, 2007, 2008 by INRIA 
 * Authors:
 *      Djame Seddah <Djame.Seddah@inria.fr>
 *      Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>
 * ----------------------------------------------------------------
 *
 *  tag_compile.pl -- TAG compiler for DyALog Compiler
 *
 * ----------------------------------------------------------------
 * Description
 *  Compile normalized TAGs into SD-2SA Meta Transitions
 *  (transitions are grouped by pairs Call/Return and continuations
 *   used in place of nabla predicates)
 *
 *  Substitution trees and substitution nodes are handled with PDA transitions
 *  (FIRST, LAST and NEXT instead of SA-FIRST, SA-LAST, SA-NEXT)
 * ----------------------------------------------------------------
 */

:-include 'header.pl'.
:-require 'tag_normalize.pl'.
:-require 'tag_maker.pl'.
:-require 'reader.pl'.

:-rec_prolog tag_compile_tree/5, tag_compile_subtree/8.

:-rec_prolog
	tag_compile_subtree_noadj/8,
	tag_compile_subtree_adj/10,
	tag_compile_subtree_potential_adj/10.
:-extensional parse_mode/1.

:-light_tabular tagguide_pred/2.
:-mode(tagguide_pred/2,+(+,-)).

:-light_tabular tagguide_name/3.
:-mode(tagguide_name/3,+(+,+,-)).

:-light_tabular verbose_tag/1.

verbose_tag(Id) :-
	recorded( verbose_tag ),
	\+ recorded( guide(Id,_,_) )
	.

:-light_tabular core_info/2.
:-mode(core_info/2,+(+,-)).

core_info(Info::tag_anchor{},Info).
core_info((XInfo * _),Info) :- core_info(XInfo,Info).
%%core_info(Info::strip(_),Info).
core_info(strip(XInfo),Info) :- core_info(XInfo,Info).

%% Imported from tig.pl
:-std_prolog make_tig_callret/7.

pgm_to_lpda( register_predicate(Pred::tag(F/N)), '*PROLOG-FIRST*'(Call) :> Cont ) :-
	(   recorded( public(Pred) ) xor recorded( public(tag('*default*')) ) ),
	(   recorded( tag_features(F,Top,Bot) )
	xor functor(Top,F,N),
	    functor(Bot,F,N)
	),
	Node = tag_node{ label => F,
			 top => Top,
			 bot => Bot,
			 adj => no,
			 kind => subst,
			 children => []
		       },
	tag_numbervars([Node,L,R,Info],Id,_),
	(   Body = Node,
	    Type=prolog
	;   
	    Body = '$answers'(Node),
	    Type=prolog
	;
	    Body = '$protect'(Node),
	    Type = protect(prolog)
	),
	Call = '$tagcall'(Body,L,R,Info),
	tag_compile_subtree(Id,Body,L,R,'*PROLOG-LAST*',Cont,Type,Info)
	.
	
pgm_to_lpda( T::tag_tree{ name=> Name, family=> Family  }, Trans ) :-
%%	format('norm ~w ->\n',[T]),
	normalize(T,L),
%%	format('\t~w\n\t~w\n',[T,L]),
	(   recorded( Info::tag_anchor{ family => Family } )
	xor Info = tag_anchor{ family => Family }
	),
	tag_numbervars(Info,Name,_),
	tag_compile_tree(Name,L,Trans1,dyalog,Info),
	install_loader(Name,
		       Loader^tag_autoloader_wrapper(Name,L,Info,Loader),
		       Trans1,
		       Trans),
%%	format('TAG Tree ~w =>\n\t ~w\n\n',[T,Trans1]),
	true
	.

tag_compile_tree(Id,Tree,Trans,Type,Info) :-
	%%	format('Tree Info ~w ~w\n',[Tree,Info]),
	(   Tree = tag_node{ label=> NT, top=> A, spine => Spine }
	xor Tree = guard{ goal => tag_node{ label => NT, top => A, spine => Spine } }
	),
	tree_info(Tree,Info),
	%%	format('\t--> ~w ~w\n',[Tree,Info]),
	tag_numbervars(Left,Id,_),
	tag_numbervars(Right,Id,_),
	functor(A,F,N),
	register_predicate(tag(F/N)),
	( Spine == yes  ->
	    \+ recorded( tig(Id,_) ), % handling of TIG tree delegated to tig.pl
	    make_tag_top_callret(A,Left,Right,Call,Return),
	    Trans = '*SA-AUX-FIRST*'(Call) :> Cont,
	    Last1 = '*SA-AUX-LAST*'(Return)
	;   
	    make_tag_subst_callret(A,Left,Right,Call,Return),
	    Trans = '*SAFIRST*'(Call) :> Cont,
	    Last1 = '*LAST*'(Return)
	),
	( verbose_tag(Id) ->
	  core_info(Info,tag_anchor{ family => HyperTag }),
	  tag_numbervars(HyperTag,Id,_),
	  body_to_lpda(Id,verbose!struct(Id,HyperTag),Last1,Last,Type)
	;   
	  Last = Last1
	),
	tag_numbervars(Tree,Id,_),
	tag_compile_subtree(Id,Tree,Left,Right,Last,Cont2,Type,Info),
	( guide(strip) ->
	    tagguide_pred(NT,SNT),
	    ( Spine == yes ->
		%% we know that this is not a TIG tree
		tagguide_name(Id,left,Left_Id),
		tag_numbervars(Foot_Left,Id,_),
		make_tig_callret(left,SNT,SNT,Left,Foot_Left,Left_Call,_),
		%% We check left guide at start
		%% right guide used after foot
		Cont3 = '*GUIDE*'(Left_Call,Left_Id,Cont2)
	    ;	
		tagguide_name(Id,no,GId),
		make_tag_subst_callret(SNT,Left,Right,GCall,_),
		Cont3 = '*GUIDE*'(GCall,GId,Cont2)
	    )
	;   recorded(debug) ->
	  body_to_lpda(Id,format('Selecting cat=~w left=~w tree ~w\n',[A,Left,Id]),Cont2,Cont3,Type)
	;
	    Cont3 = Cont2
	),
	( recorded( tagfilter(Id^Left^Right^NT^A^TagFilter) ),
	  recorded( lctag(Id,_,_,_) )
	  ->
	  tag_numbervars(TagFilter,Id,_),
	  body_to_lpda(Id, TagFilter, Cont3,Cont,Type)
	;
	  Cont=Cont3
	),
%%	format('TAG Tree ~w =>\n\t ~w\n',[Tree,Trans]),
	true
	.

tag_compile_subtree(Id,T,Left,Right,Cont,Trans,Type,Info):-
%%	format('TAG Subtree ~w\n',[T]),
	(   tag_compile_subtree_handler(Id,T,Left,Right,Cont,Trans,Type,Info) ->
	    true
	%% Last children in a children list
	;   T = [A] ->
	    tag_compile_subtree(Id,A,Left,Right,Cont,Trans,Type,Info)
	%% Traversing a children list
	;   T = [A|B] ->
	    tag_numbervars(Middle,Id,_),
	    tag_compile_subtree(Id,B,Middle,Right,Cont,ContB,Type,Info),
	    tag_compile_subtree(Id,A,Left,Middle,ContB,Trans,Type,Info)
	;   tag_compile_node(Id,T,Left,Right,Cont,Trans,Type,Info)
	),
%%	format('===> TAG Subtree ~w Trans=~w\n',[T,Trans]),
	true
	.

%%! @predicate tag_compile_subtree_handler(Id,T,Left,Right,Cont,Type,Info)
%%! @brief to express handlers for all cases of TAG "nodes"

:-rec_prolog tag_compile_subtree_handler/8.

%%! @case Disjunction
tag_compile_subtree_handler(Id,(A;B),Left,Right,Cont,Trans,Type,Info) :-
	handle_choice(Cont,Trans,Last,ContA,ContB),
	tag_compile_subtree(Id,A,Left,Right,Last,ContA,Type,Info),
	tag_compile_subtree(Id,B,Left,Right,Last,ContB,Type,Info)
	.

%%! @case Empty Scan node
tag_compile_subtree_handler(Id,
			    tag_node{ label => [], kind => scan },
			    Left,Right,Cont,unify(Left,Right) :> Cont,Type,Info).

%%! @case Scan node
tag_compile_subtree_handler(Id,
			    tag_node{ label => L::[_|_], kind => scan, id => Node },
			    Left,Right,Cont,Trans,Type,Info) :-
	( verbose_tag(Id)  ->
	    tag_numbervars([Cat,Lex],Id,_),
	    body_to_lpda(Id,verbose!lexical(L,Left,Right,Cat,Lex),Cont,Cont2,Type)
	;   
	    Cont2=Cont
	),
%%	format('Try Lex compile ~w in ~w\n',[L,Id]),
	scan_to_lpda(Id,L,Left,Right,Cont2,Trans1,Type),
%%	format('Done Lex compile ~w in ~w\n',[L,Id]),
	(   core_info(Info,CInfo),
	    CInfo = tag_anchor{ family => Family, coanchors => VarCoanchors },
	    recorded( tag_coanchor(Family,Node,LemmaVar, VarCoanchors) ) ->
	    tag_numbervars(LemmaVar,Id,_),
	    body_to_lpda(Id,check_coanchor(LemmaVar,Left,Right),Trans1,Trans,Type)
	;   
	    Trans=Trans1
	)
	.

%%! @case Escape node
tag_compile_subtree_handler(Id,
			    tag_node{ label => A, kind => escape },
			    Left,Right,Cont,unify(Left,Right) :> Trans,Type,Info) :-
	body_to_lpda(Id,A,Cont,Trans,Type)
	.

%%! @case Position node
tag_compile_subtree_handler(Id,
			    '$pos'(Pos),
			    Left,Right,Cont,
			    unify(Pos,Left) :> unify(Left,Right) :> Cont,
			    Type,Info
			   )
.

%%! @case Answer
tag_compile_subtree_handler(Id,'$answers'(A),Left,Right,Cont,Trans,Type,Info) :-
	tag_compile_subtree(Id,A,Left,Right,Cont,Trans,answer(Type),Info)
	.

%%! @case Protect
tag_compile_subtree_handler(Id,'$protect'(A),Left,Right,Cont,Trans,Type,Info) :-
	tag_compile_subtree(Id,A,Left,Right,Cont,Trans,protect(Type),Info)
	.


%%! @case MetaCall
tag_compile_subtree_handler(Id,
			    tag_node{ id=>NId,
				      label => '$call'(A),
				      kind => Kind,
				      top => Top,
				      bot => Bot,
				      adj => Adj,
				      children => Children,
				      spine => Spine,
				      token => Token
				    },
			    Left,Right,Cont,Trans,Type,Info) :-
	metacall_initialize,
	Node = tag_node{ id=>NId,
			 label => A,
			 kind => Kind,
			 top => Top,
			 bot => Bot,
			 adj => Adj,
			 children => Children,
			 spine => Spine,
			 token => Token
		       },
	( Type = answer(Type2) ->
	    body_to_lpda(Id,call!tagcall('$answers'(Node),Left,Right,Info),Cont,Trans,Type2)
	;  Type = protect(Type2) -> 
	    body_to_lpda(Id,call!tagcall('$protect'(Node),Left,Right,Info),Cont,Trans,Type2)
	;   
	    body_to_lpda(Id,call!tagcall(Node,Left,Right,Info),Cont,Trans,Type)
	)
	.

%%! @case Guarded Node
tag_compile_subtree_handler(Id,
			    guard{ goal => A,
				   plus => Plus,
				   minus => Minus
				 },
			    Left, Right, Cont, Trans, Type, Info) :-
	( Plus == true, Minus = fail ->
	    %% actually an unguarded node !
	    tag_compile_subtree(Id,A,Left,Right,Cont,Trans,Type,Info)
	;   Minus == fail ->
	    %% no negative guard
	  ( %% recorded( tig(Id,left) ), tree_foot(A,_) ->
	    true ->
	    tag_compile_subtree(Id,A,Left,Right,Cont,Trans1,Type,Info),
	    body_to_lpda(Id,Plus,Trans1,Trans,Type)
	  ;
	    body_to_lpda(Id,Plus,Cont,Trans1,Type),
	    tag_compile_subtree(Id,A,Left,Right,Trans1,Trans,Type,Info)
	  )
	;   handle_choice(Cont,Trans,Last,ContA,ContB),
	  body_to_lpda(Id,Minus,unify(Left,Right) :> Last,ContB,Type),
	  ( %% recorded( tig(Id,left) ), tree_foot(A,_) ->
	    true ->
	    tag_compile_subtree(Id,A,Left,Right,Last,ContA1,Type,Info),
	    body_to_lpda(Id,Plus,ContA1,ContA,Type)
	  ;
	    body_to_lpda(Id,Plus,Last,ContA1,Type),
	    tag_compile_subtree(Id,A,Left,Right,ContA1,ContA,Type,Info)
	  )
	)
	.

%%! @case Kleene star
%% only on non spine nodes
tag_compile_subtree_handler(Id,
			    KleeneStruct:: @*{ goal => A,
					       vars => Vars,
					       from => From,
					       to => To,
					       collect_first=>UserFirst,
					       collect_loop=>UserParam,
					       collect_next=>UserArg,
					       collect_last=>UserLast
					     },			    
			    Left,Right,Cont,Trans,Type,Info) :-
%%	format('Compiling TAG Kleene\n',[A]),
	update_counter(kleene, KleeneId),
        name_builder('_kleene~w',[KleeneId],KleeneLabel),
	kleene_conditions(KleeneCond::[From,To],[Counter,CounterInc],InitCond,ExitCond,LoopCond),
	tag_numbervars([Left2,Middle,Right],Id,_),
	tag_numbervars([Counter,CounterInc,InitCond,ExitCond,LoopCond,KleeneStruct],Id,_),
	Args1 = [Counter,Left2],
	Loop_Args1 = [CounterInc,Middle],
	append(Args1,UserParam,Args),
	append(Loop_Args1,UserArg,Loop_Args),
	tupple(Vars,TA),
	tupple(Cont,TCont),
	extend_tupple([Right,UserLast,KleeneCond],TCont,RTCont),
	body_to_lpda( Id,
		      (	  Left2=Right, ExitCond, UserLast = UserParam),
		      Cont,
		      ContExit,
		      Type
		    ),
	tag_compile_subtree(Id,A,
			    Left2,Middle,
			    '*KLEENE-LAST*'(KleeneLabel,
					    Loop_Args,
					    Args,
					    TA,
					    RTCont,
					    noop
					   ),
			    ContLoop2,
			    Type,
			    Info),
	body_to_lpda(Id,LoopCond,ContLoop2,ContLoop,Type),
	Trans1 = '*KLEENE*'(KleeneLabel,
			    TA,
			    '*KLEENE-ALTERNATIVE*'(ContExit,ContLoop)
			    ),
	body_to_lpda(Id,(InitCond,Left=Left2,UserFirst=UserParam),
		     Trans1,
		     Trans,
		     Type),
%%	format('--> Kleene Trans ~w\n',[Trans]),
	true
	.

:-rec_prolog tag_compile_node/8.

tag_compile_node(Id,
                 T,
		 Left,
		 Right,
		 Cont,
		 Trans,
		 Type,
		 Info
		) :-
	(   %% Foot node
	    T = tag_node{ label => NT, bot => A, kind => foot } ->
	    \+ recorded( tig(Id,_) ), % special TIG case for foot
	    tag_numbervars(A,Id,_),
	    functor(A,F,N),
	    Pred = tag_foot(F/N),
	    wrapping_predicate(Pred),
	    decompose_args(Pred,[Left,Right,A],Args),
	    Trans = '*SA-FOOT-WRAPPER-CALL*'(Pred,Args,Cont2),
	    ( fail, guide(strip) ->
		tagguide_pred(NT,SNT),
		tagguide_name(Id,right,GId),
		tag_numbervars(_Right,Id,_),
		make_tig_callret(right,SNT,SNT,Right,_Right,GCall,_),
		Cont2 = '*GUIDE*'(GCall,GId,Cont)
	    ;
		Cont2 = Cont
	    )
	%% SubstFoot Node (same as subst but keeps spine)
	;   T = tag_node{ bot => A, kind => substfoot } ->
	    tag_numbervars(A,Id,_),
	    make_tag_bot_callret(A,Left,Right,Call,Return),
	    Trans = '*SA-SUBSTFOOT*'(Call,Return,Cont)
	%% Node without adjonction
	;   T = tag_node{ adj => no } ->
	    tag_compile_subtree_noadj(Id,T,Left,Right,Cont,Trans,Type,Info)
	%% Node with full adjonction
	%%	;   allowed_wrap_adj(Id,T) ->
	%%    tag_compile_subtree_adj(Id,T,Left,Right,Cont,Trans,Type,Info)
	%% Node with potential tig adj but no wrapped adj (try left first)
	;   T = tag_node{ top=>Top },
	    tig_compile_subtree_left(Id,T,Left,Right,Cont,Trans,Type,Info,Top),
	    true
	)
	.


:-rec_prolog tig_compile_subtree_left/9.

:-std_prolog allowed_wrap_adj/2.

allowed_wrap_adj(Id,T::tag_node{ spine => Spine, bot => A }) :-
%%	format('Check allowed adj ~w ~w\n',[Id,T]),
	\+ ( Spine = yes, recorded(tig(Id,_))),
	functor(A,F,N),
	%% WARNING:
	%% As soon as some tig information is provided for some predicate
	%% we assume that we are using TIG
	%% => adj on F/N is possible if and only if adjkind(F/N,wrap) provided !
	(   recorded(adjkind(F/N,wrap)) xor
	(   \+ recorded(adjkind(F/N,_)),
	    \+ recorded(adjkind(_,no))
	)
	),
%%	format('--> Allowed adj\n',[]),
	true
	.

decompose_args( tag_foot(F/N), L::[Left,Right,T], Args ) :-
	functor(T,F,N),
	( N > 30 -> Args = L ; T =.. [_|ArgsT], Args = [Left,Right|ArgsT] )
	.

wrapping_predicate( Pred::tag_foot(F/N) ) :-
	decompose_args(Pred,[Left,Right,A],Args1),
	make_tag_bot_callret(A,Left,Right,Call,Return),
	tag_numbervars(Args::[Closure|Args1],Id,_),
	record( '*SA-FOOT-WRAPPER*'(Pred,
				    Id,
				    Args,
				    '*SA-FOOT*'(Call,Return,'*SA-PROLOG-LAST*')
				   )
	      )
	.

:-std_prolog tag_compile_info_anchor/2.

tag_compile_info_anchor( Info::tag_anchor{ family => Family,
					   equations => VarEq
					 },
			 T::tag_node{ id => Node, bot=>Bot, top=>Top }
		       ) :-
	(   recorded( tag_equation(Family,Node,Top,Bot, VarEq ) ) xor true )
	.

tag_compile_subtree_potential_adj(Id,
				  T::tag_node{ top => TrueTop,
					       bot => TrueBot,
					       id => A_Id
					     },
				  Left,
				  Right,
				  Cont,
				  Trans,
				  Type,
				  Info,
				  Top,
				  Bot
				 ) :-
%%	format('Try Potential Adj ~w ~w\n',[A_Id,Id]),		
	(   allowed_wrap_adj(Id,T) ->
%%	    format('Potential HERE1\n',[]),
	    tag_compile_subtree_adj(Id,T,Left,Right,Cont,Trans,Type,Info,Top,Bot)
	;
%%	    format('Potential HERE2\n',[]),
	    ( Top == TrueTop, Bot==TrueBot ->
	      tag_numbervars(Top,Id,_),
	      tag_numbervars(Bot,Id,_),
%%            Trans = unify(Top,Bot) :> Trans1
              destructure_unify(Top,Bot,Trans1,Trans)
	    ; Top=Bot ->
	      tag_numbervars(Top,Id,_),
	      Trans = Trans1
	    ;
	      tag_numbervars(Top,Id,_),
	      tag_numbervars(Bot,Id,_),
	      %%            Trans = unify(Top,Bot) :> Trans1
              destructure_unify(Top,Bot,Trans1,Trans)
	    ),
%%	    format('Potential HERE3\n',[]),
	    tag_compile_subtree_noadj(Id,T,Left,Right,Cont,Trans1,Type,Info)
	)
	.

tag_compile_subtree_adj(Id,
			T::tag_node{ label => NT,
				     top => TrueTop,
				     bot=> TrueBot,
				     adj=>Adj,
				     id=>A_Id ,
				     spine=>Spine},
			Left,
			Right,
			Cont,
			'*SA-ADJ-WRAPPER-CALL*'(Pred,Args,BCont,Below,RCont),
			Type,
			Info,
			Top,
			Bot
		       ) :-
	tag_numbervars(Foot_Left,Id,_),
	tag_numbervars(Foot_Right,Id,_),
	tag_numbervars(Top,Id,_),
	tag_numbervars(Bot,Id,_),
	tag_numbervars(BCont,Id,_),
	tag_numbervars(RCont,Id,_),
	%%
	functor(Top,F,N),
	(   Adj == strict,
	    \+ ( recorded(adjkind(F/N,Mode)), Mode \== wrap )
	->  %% Mandatory adjonction and no concurrent tig adj
	    Pred = tag_adj_strict(F/N)
	;   %% Non mandatory adjonction or concurrent tig adj
	    Pred = tag_adj(F/N)
	),
	wrapping_predicate(Pred),
	( recorded( tag_features(NT,Top,Bot) ) xor true),
	decompose_args(tag_adj(F/N),
		       [Left,Right,Foot_Left,Foot_Right,Top,Bot,A_Id],
		       Args),
	tag_compile_subtree_noadj(Id,
				  T,
				  Foot_Left,
				  Foot_Right,
				  '*SA-PROLOG-LAST*'(Cont),
				  Below,
				  Type,
				  Info
				 )
	.
/*
wrapping_predicate( Pred::tag_adj(F/N)) :-
	decompose_args( Pred,
			[Left,Right,Foot_Left,Foot_Right,Top,Bot,A_Id],
			Args1
		      ),
	make_tag_top_callret(Top,Left,Right,Call,Return),
	make_tag_bot_callret(Bot,Foot_Left,Foot_Right,Foot_Call,Foot_Return),
	tag_numbervars(Args::[BClosure,RClosure,BCont|Args1],_,_),
	record( '*SA-ADJ-WRAPPER*'( Pred,
				    Args,
				    ['*SA-NOADJ-FIRST*'(Call,Foot_Call,'*SA-PROLOG-LAST*')
				    |'*SA-ADJ-FIRST*'(Call,Foot_Call,'*SA-PROLOG-LAST*')],
				    ['*SA-NOADJ-LAST*'(Return,Foot_Return,'*SA-PROLOG-LAST*',A_Id)
				    |'*SA-ADJ-LAST*'(Return,Foot_Return,'*SA-PROLOG-LAST*',A_Id)]
				  ) )
	.
*/

wrapping_predicate( Pred::tag_adj(F/N)) :-
	decompose_args( Pred,
			[Left,Right,Foot_Left,Foot_Right,Top,Bot,A_Id],
			Args1
		      ),
	make_tag_top_callret(Top,Left,Right,Call,Return),
	make_tag_bot_callret(Bot,Foot_Left,Foot_Right,Foot_Call,Foot_Return),
	tag_numbervars(Args::[BClosure,RClosure,BCont|Args1],Id,_),
	body_to_lpda(Id,
		     (Call=Foot_Call,Return=Foot_Return),
		     '*SA-PROLOG-LAST*',
		     NoAdjCont,
		     dyalog),
	_AdjCont = '*SA-ADJ-FIRST*'(Call,Foot_Call,'*SA-PROLOG-LAST*'),
	( recorded(tag_features(NT,Top,Bot)) xor NT=F),
	( recorded( tagfilter_cat(NT^wrap^Left^Right^Filter) )
	->
	  tag_numbervars(Filter,Id,_),
	  %%	  format('Compiling ~w ~w ~w with ~w\n',[F,Mode,dyalog,Filter]),
	  body_to_lpda(Id,Filter,_AdjCont,AdjCont,dyalog),
	  %%	  format('==> ~w\n',[ContAdj]),
	  true
	;
	  AdjCont = _AdjCont
	),
	record( '*SA-ADJ-WRAPPER*'( Pred,
				    Args,
				    [ NoAdjCont | AdjCont],
				    ['*SA-PROLOG-LAST*'
				    |'*SA-ADJ-LAST*'(Return,Foot_Return,'*SA-PROLOG-LAST*',A_Id)]
				  ) )
	.


wrapping_predicate( Pred::tag_adj_strict(F/N)) :-
	decompose_args( tag_adj(F/N),
			[Left,Right,Foot_Left,Foot_Right,Top,Bot,A_Id],
			Args1
		      ),
	make_tag_top_callret(Top,Left,Right,Call,Return),
	make_tag_bot_callret(Bot,Foot_Left,Foot_Right,Foot_Call,Foot_Return),
	tag_numbervars(Args::[BClosure,RClosure,BCont|Args1],Id,_),
	_AdjCont = '*SA-ADJ-FIRST*'(Call,Foot_Call,'*SA-PROLOG-LAST*'),
	( recorded(tag_features(NT,Top,Bot)) xor NT=F),
	( recorded( tagfilter_cat(NT^wrap^Left^Right^Filter) )
	->
	  tag_numbervars(Filter,Id,_),
	  %%	  format('Compiling ~w ~w ~w with ~w\n',[F,Mode,dyalog,Filter]),
	  body_to_lpda(Id,Filter,_AdjCont,AdjCont,dyalog),
	  %%	  format('==> ~w\n',[ContAdj]),
	  true
	;
	  AdjCont = _AdjCont
	),
	record( '*SA-ADJ-WRAPPER*'( Pred,
				    Args,
				    AdjCont,
				    '*SA-ADJ-LAST*'(Return,Foot_Return,'*SA-PROLOG-LAST*',A_Id)
				  )
	      )
	.


:-rec_prolog optimize_topbot_args/4.

optimize_topbot_args([],[],L,L).
optimize_topbot_args([T|Top],[B|Bot],In,Out) :-
	optimize_topbot_args(Top,Bot,In,M),
	( T == B ->
	  Out = [T|M]
	;
	  Out = [T,B|M]
	).

decompose_args(tag_adj(F/N),
	       L::[Left,Right,Foot_Left,Foot_Right,Top,Bot,A_Id],
	       Args
	      ) :-
	functor(Top,F,N),
	functor(Bot,F,N),
	( recorded(tag_features(_,Top,Bot)) ->
	  Top =.. [F|TopArgs],
	  Bot =.. [F|BotArgs],
	  optimize_topbot_args(TopArgs,BotArgs,[A_Id],TopBotArgs),
	  Args = [Left,Right,Foot_Left,Foot_Right|TopBotArgs]
	; N > 15 ->
	  Args = L 
	;
	  make_tag_top_callret(Top,Left,Right,Call,Return),
	  make_tag_bot_callret(Bot,Foot_Left,Foot_Right,Foot_Call,Foot_Return),
	  Call =.. [_|Call_Args],
	  Foot_Call =.. [_|Foot_Call_Args],
	  append(Call_Args,Foot_Call_Args,Adj_First_Args),
	  Return =.. [_|Return_Args],
	  Foot_Return =.. [_|Foot_Return_Args],
	  append([A_Id|Return_Args],Foot_Return_Args,Adj_Last_Args),
	  append(Adj_First_Args,Adj_Last_Args,Args)
	),
%%	format('decompose_adj_args ~w => ~w ~w\n',[F/N,Adj_First_Args,Adj_Last_Args]),
	true
	.

:-light_tabular tag_custom_subst/2.
:-mode(tag_custom_subst/2,+(+,-)).

tag_custom_subst(Type,Pred) :-
	(\+ Type = protect(_)),
	recorded( tag_custom_subst(Pred) ),
	true
	.


:-rec_prolog tag_wrap_skipper/8.
%%:-std_prolog tag_wrap_skipper/6.

tag_wrap_skipper(Id,Left,_Left,Right,G,Cont,Trans,Type) :-
	( recorded( skipper(_,_,_,_) ) ->
	    tag_numbervars([_Left],Id,_),
	    body_to_lpda(Id,G,Cont,Cont1,Type),
	    ( verbose_tag(Id) ->
		Trace = true
	    ;
		Trace = false
	    ),
	    wrapping_predicate(Pred::tag_skipper(Trace)),
	    decompose_args(tag_skipper,[Left,_Left], Args),
	    Trans = '*WRAPPER-CALL*'(Pred,Args,Cont1),
	    true
	;   
	    Left = _Left,
	    body_to_lpda(Id,G,Cont,Trans,Type)
	)
	.


%% Inspired of multiple tig adjoining in tig.pl
%% a direct coding with @* doesn't seem to work (need to investigate)
wrapping_predicate( Pred::tag_skipper(TraceFlag) ) :-
	decompose_args( tag_skipper, [Left,Right], Args1),
	decompose_args( tag_skipper, [Middle,Right], Args2),
	recorded( skipper(Left,_A,Middle,Skip) ),
	tag_numbervars(Args::[Closure|Args1],Id,_),
	tag_numbervars(Args2,Id,_),
	( TraceFlag = true ->
	    tag_numbervars(Lex,Id,_),
	    Trace = '$tagop'(skip,verbose!epsilon(_A,Left,Middle,Lex)),
	    SkipTrace = (Skip,Trace)
	;   
	    SkipTrace = Skip
	),
	tag_numbervars([SkipTrace],Id,_),
	Inner_Call = '*WRAPPER-INNER-CALL*'(tag_skipper(TraceFlag),Args2,'*PROLOG-LAST*'),
	body_to_lpda(Id,SkipTrace,Inner_Call,InnerCont,dyalog),
%%	ContSkip = '*CONT*'(InnerCont),
	ContSkip = InnerCont,
	ContNoSkip = unify(Left,Right) :> '*PROLOG-LAST*' :> fail,
	Cont = '*ALTERNATIVE-LAYER*'(ContNoSkip,ContSkip),
	%%		body_to_lpda(Id,Wrapper,'*PROLOG-LAST*' :> fail,Cont,dyalog),
	%%	Cont = unify(Left,Right) :> '*PROLOG-LAST*',
	record( Wrap::'*WRAPPER*'(Pred,Args,Cont) ),
	true
	.

decompose_args( tag_skipper, L, L ).

tag_compile_subtree_noadj(Id,
			  T::tag_node{ id => Node,
				       bot => A,
				       kind => Kind,
				       children=>Children,
				       label=>Label,
				       token=> Token
				     },
			  Left,
			  Right,
			  Cont,
			  Trans,
			  Type,
			  Info
			 ):-
	tag_numbervars(A,Id,_),
	%% substitution node
	%% => node not on spine
	( Kind == subst ->
	    (	Type = answer(_) ->
		make_tag_subst_callret(A,Left,Right,Call,Return),
		Trans1='*SA-PSEUDO-SUBST*'(Call,Return,Cont,Node)
	    ;	tag_custom_subst(Type,Custom_Subst) ->
		K =.. [Custom_Subst,Label,A,Left,Right,Node],
		body_to_lpda(Id,K,Cont,Trans1,Type) 
	    ;	%% fail ->
		functor(A,F,N),
		decompose_args(tag_subst(F/N), [Left,Right,Node,A], Args ),
		wrapping_predicate(tag_subst(F/N,Label)),
		Trans1='*WRAPPER-CALL*'(tag_subst(F/N),Args,Cont)
	    /* Nonwrapped version: to test efficiency
	    ;	make_tag_subst_callret(A,Left,Right,Call,Return),
		Trans1='*SA-SUBST*'(Call,Return,Cont,Node)
	    */
	    ),
	    %%	    Trans1 = '*SA-SUBST*'(Call,Return, Cont,Node),
	    (	core_info(Info,CInfo),
		CInfo = tag_anchor{ family => Family, coanchors => VarCoanchors },
		recorded( tag_coanchor(Family,Node,LemmaVar, VarCoanchors) ) ->
		tag_numbervars(LemmaVar,Id,_),
		body_to_lpda(Id,check_coanchor(LemmaVar,Left,Right),Trans1,Trans,Type)
	    ;	
		Trans=Trans1
	    )
	%% Coanchor node
	%% => node not on spine
	;   Kind == coanchor ->
	    %%	    tag_numbervars(Token,Id,_),
%%	    format( 'CoAnchor ~w ~w -> ~w\n',[Id,Token,Info]),
	    Coanchor = coanchor(Token,Label,_Left,Right,A),
	    ( verbose_tag(Id) ->
		tag_numbervars(Lex,Id,_),
		Trace = '$tagop'(Node,verbose!coanchor(Token,_Left,Right,A,Lex)),
		CoanchorTrace = ( Coanchor , Trace )
	    ;	
		CoanchorTrace = Coanchor
	    ),
	    tag_wrap_skipper(Id,Left,_Left,Right,
			     CoanchorTrace,
			     Cont,
			     Trans1,
			     Type ),
%%	    format('Here.2 Wrapper=~w\n\tCont2=~w\n\tType=~w\n\tInfo=~w\n',[Wrapper,Cont2,Type,Info]),
%%	    format('Here.3\n',[]),
	    (	core_info(Info,CInfo),
		CInfo = tag_anchor{ family => Family, coanchors => VarCoanchors },
		recorded( tag_coanchor(Family,Node,LemmaVar, VarCoanchors) ) ->
		tag_numbervars(LemmaVar,Id,_),
		body_to_lpda(Id,check_coanchor(LemmaVar,_Left,Right),Trans1,Trans,Type)
	    ;	
		Trans=Trans1
	    )
	%% Anchor node
	%% => node not on spine
	;   Kind == anchor ->
	    %% tag_numbervars(Token,Id,_),
	    %%	    format( 'Anchor ~w ~w -> ~w\n',[Id,Token,Info]),
	    Anchor = anchor(Info1,Token,Label,_Left,Right,A),
	    ( verbose_tag(Id) ->
		tag_numbervars(Lex,Id,_),
		Trace = '$tagop'(Node,verbose!anchor(Token,_Left,Right,Id,A,Lex,Info1)),
		AnchorTrace = (Anchor,Trace)
	    ;	
		AnchorTrace = Anchor
	    ),
	    core_info(Info,Info1),
	    tag_wrap_skipper(Id,Left,_Left,Right,
			     AnchorTrace,
			     Cont,
			     Trans,
			     Type),
	    %%	    Info = tag_anchor{ family => Family },
	    true
	%% Inside node
	%% nothing to do for this node, except compiling its children
	;   Kind ==  std ->
	    tag_compile_subtree(Id,Children,Left,Right,Cont,Trans,Type,Info)
	;   
	    error('TAG to SA compilation error: ~w',[T])
	)	,
%%	format('\n---------------\nTAG NoAdj ~w\n~K\n----------------\n',[T,Trans]),
	true
	.

decompose_args( tag_subst(F/N), [Left,Right,Node,T], Args ) :-
	functor(T,F,N),
	( N > 30 -> Args = [Left,Right,Node,T] ; T =.. [_|ArgsT], Args = [Left,Right,Node|ArgsT] )
	.

wrapping_predicate( tag_subst(F/N,NT) ) :-
	Pred = tag_subst(F/N),
	decompose_args(Pred,[Left,Right,Node,A],Args1),
	make_tag_subst_callret(A,Left,Right,Call,Return),
	tag_numbervars(Args::[Closure|Args1],Id,_),
	_Cont = '*SA-SUBST*'(Call,Return,'*PROLOG-LAST*',Node),
	( recorded( tagfilter_cat(NT^subst^Left^Right^Filter) )
	->
	  tag_numbervars(Filter,Id,_),
	  %%	  format('Compiling ~w ~w ~w with ~w\n',[F,Mode,dyalog,Filter]),
	  body_to_lpda(Id,Filter,_Cont,Cont,dyalog),
	  %%	  format('==> ~w\n',[ContAdj]),
	  true
	;
	  Cont = _Cont
	),
	record( '*WRAPPER*'( Pred,
			     Args,
			     Cont
			   ) )
	.

:-std_prolog tag_numbervars/3.

tag_numbervars(T,Id,M) :-
        ( var(Id) ->
            gensym(Id),
	    numbervars(T,1,M),
	    mutable(MM,M),
            record( numbervars(Id,MM) )
        ;   recorded( numbervars(Id,MM) ) ->
	    mutable_read(MM,N),
            numbervars(T,N,M),
	    mutable(MM,M)
	;   numbervars(T,1,M),	% use name of tree as id to number variables
	    mutable(MM,M),
            record( numbervars(Id,MM) )
        )
        .

:-std_prolog numbervars_delete/1.

numbervars_delete(Id) :- erase(numbervars(Id,_)).

%% Imported from tagguide.pl
:-std_prolog tagguide_strip/3.

tag_normalize_and_compile(Id,Tree,Left,Right,Cont,Trans,Type) :-
	normalize(tag_tree{ tree=> (tree Tree) },T),
	tag_numbervars(T,Id,_),
	tag_compile_subtree(Id,T,Left,Right,Cont,Trans1,Type,[]),
	( guide(strip) ->
	    tagguide_strip(T,no,ST),
	    tag_compile_subtree(Id,ST,Left,Right,fail,ContGuide,Type,[]),
	    Trans = '*WAIT*'(Trans1) :> ContGuide
	;   
	    Trans = Trans1
	),
	true
	.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Compilation of TAG lemma

pgm_to_lpda( T::tag_lemma(Lemma,Label,Anchors), Trans ) :-
	Clause = (normalized_tag_lemma(Lemma,Label,_Family,_VarCoanchors,_VarEquations) :- _Body ),
	( var(Anchors) ->
	    error( 'tag_lemma: unexpected variable as anchor ~w', [T] )
	;   Anchors = tag_anchor{} ->
	    tag_compile_lemma_anchor( Anchors, Clause )
	;   domain( Anchor, Anchors ),
	    tag_compile_lemma_anchor( Anchor, Clause )
	),
	%% format( 'LEMMA ~w\n\n', [Clause] ),
	clause_to_lpda( Clause, Trans )
	.

:-std_prolog tag_compile_lemma_anchor/2.

tag_compile_lemma_anchor( T::tag_anchor{ family => Family,
					 coanchors => Coanchors,
					 equations => Equations
				       },
			  Clause
			) :-

	( var(Family) ->
	    error( 'tag_lemma: unexpected variable as family ~w', [T] )
	;   atom(Family) ->
	    (	recorded( Info::tag_anchor{ family => Family,
					    coanchors => VarCoanchors,
					    equations => VarEquations
					  } )
	    ->	true
	    ;
		warning( 'tag_lemma: reference to missing anchor model ~w', [Family] ),
		VarCoanchors=[],
		VarEquations=[]
	    ),
	    Coanchors ?= [],
	    Equations ?= [],
	    Clause = (normalized_tag_lemma(_,_,Family,VarCoanchors,VarEquations) :- Body ),
	    tag_compile_lemma_coanchors( Coanchors, VarCoanchors, Family, Body),
	    tag_compile_lemma_equations( Equations, VarEquations, Family)
	;   domain(F,Family),
	    tag_compile_lemma_anchor( tag_anchor{ family => F,
						  coanchors => Coanchors,
						  equations => Equations
						},
				      Clause
				    )
	)
	.

:-rec_prolog tag_compile_lemma_coanchors/4.

tag_compile_lemma_coanchors( [], _, _, true ).
tag_compile_lemma_coanchors( [Node=Lemma|Coanchors],
			     VarCoanchors,
			     Family,
			     (	 check_coanchor_in_anchor(LemmaVar,LemmaList),
				 Body
			     )
			   ) :-
	( recorded( tag_coanchor(Family,Node,LemmaVar,VarCoanchors) ) ->
	    tag_compile_lemma_check(Lemma,LemmaList),
	    tag_compile_lemma_coanchors(Coanchors,VarCoanchors,Family,Body)
	;   
	    error( 'Node ~w not referenced as a coanchor for family ~w\n',[Node,Family] )
	)
	.

:-std_prolog tag_compile_lemma_check/2.

tag_compile_lemma_check(Lemma,LemmaList) :-
	( Lemma = (L1 ; L2) ->
	    LemmaList = [L1|LemmaList2],
	    %%	    tag_compile_lemma_check(L1,LemmaVar,Body1),
	    tag_compile_lemma_check(L2,LemmaList2)
	;   
	    LemmaList = [Lemma]
	)
	.

:-rec_prolog tag_compile_lemma_equations/3.

:-std_prolog tag_anchor_analyse_equation/3.  %% See in reader.pl

tag_compile_lemma_equations([],_,_).
tag_compile_lemma_equations([E|Equations],VarEquations,Family) :-
	( var(E) ->
	    error( 'tag_lemma: unexpected variables in equation', [] )
	;   E = (U at Node) ->
	    (	recorded( tag_equation(Family,Node,Top,Bot,VarEquations) ) ->
		tag_anchor_analyse_equation(U,Top,Bot),
		tag_compile_lemma_equations(Equations,VarEquations,Family)
	    ;	
		error( 'Node ~w not referenced to support equations for family ~w\n',[Node,Family] )
	    )
	;   
	    error( 'tag_lemma: bad format for equation ~w', [E] )
	)
	.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

:-std_prolog tree_info/2.

%% Assume Info is a core info of the form tag_anchor{} or strip(tag_anchor{})
tree_info(T,Info) :-
	( var(T) -> true
	;   T=[] -> true
	;   T=[A|B] ->
	  tree_info(A,Info),
	  tree_info(B,Info)
	;   T = @*{ goal => A } ->
	  tree_info(A,Info)
	;   T = (A ## B) ->
	  tree_info(A,Info),
	  tree_info(B,Info)
	;   T = (A,B) ->
	  tree_info(A,Info),
	  tree_info(B,Info)
	;   T= (A;B) ->
	  tree_info(A,Info),
	  tree_info(B,Info)
	;   T = '$answers'(A) ->
	  tree_info(A,Info)
	;   T = guard{ goal => A } ->
	  tree_info(A,Info)
	;   T = tag_node{ kind=>Kind, children=>Children } ->
	  (	domain(Kind,[scan,escape])
	    xor ( Info = strip(_) ;
		    tag_compile_info_anchor(Info,T)),
		tree_info(Children,Info))
	;   true
	)
	.

:-std_prolog tree_foot/2.

%% Assume Info is a core info of the form tag_anchor{} or strip(tag_anchor{})
tree_foot(T,Foot) :-
	( var(T) -> fail
	;   T=[] -> fail
	;   T=[A|B] -> fail
	;   T = @*{ goal => A } ->
	    tree_foot(A,Foot)
	;   T = (A ## B) ->
	    (	tree_foot(A,Foot) xor    tree_foot(B,Foot) )
	;   T= (A;B) ->
	    (	tree_foot(A,Foot) xor    tree_foot(B,Foot)  )
	;   T= (A,B) ->
	    (	tree_foot(A,Foot) xor    tree_foot(B,Foot)  )
	;   T = '$answers'(A) ->
	    tree_foot(A,Foot)
	;   T= tag_node{ kind => foot } -> Foot = T
	;   T = tag_node{ kind=>Kind, children=>Children } ->
	    domain(T1,Children),
	    tree_foot(T1,Foot)
	;   T = guard{ goal => T1 } ->
	    tree_foot(T1,Foot)
	;   fail
	)
	.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-std_prolog tag_autoloader_wrapper/4.

tag_autoloader_wrapper(Name,L,Info,Loader) :-
	tag_autoloader(Name,L::tag_node{ spine => Spine,
					 label => Cat},Info,Loader1),
	( Spine \== yes ->
	  Mode = subst
	; recorded(tig(Name,Mode)) ->
	  true
	; Mode = wrap
	),
	mutable(LCM,true),
	every(( recorded(D::lctag(Name,Kind,Val,Trees)),
		mutable_read(LCM,_LCS),
		_LC=( record(K::lctag(Kind,Val,Name,Trees)) ),
		tag_numbervars(_LC,Name,_),
		mutable(LCM,(_LC,_LCS))
	      )),
	mutable_read(LCM,LCS),
	LCS2 = ( record(loaded_tree(Name)),
		 record(lctag_map(Cat,Mode,Name)),
		 LCS ),
	( \+ recorded(lctag(Name,_,_,_)) ->
	  LCS3= ( record(lctag_ok(Name)), LCS2 )
	;
	  LCS3 = LCS2
	),
	(LCS3 = true ->
	 Loader = Loader1
	;
	 Loader = (Loader1 -> LCS3 ; fail)
	),
%%	format('LCS loader ~w ~w\n',[Name,Loader]),
	true
	.


:-std_prolog tag_autoloader/4.
%% +pattern: tag_autoloader(Tree,Info,Loader)
%% +desc: extract the autoload part of a TAG tree

%% Assume Info is a core info of the form tag_anchor{}
tag_autoloader(Name,T,Info,Loader) :-
	( T=[A] ->
	    tag_autoloader(Name,A,Info,Loader)
	;   T=[A|B] ->
	    tag_autoloader(Name,A,Info,Loader1),
	    tag_autoloader(Name,B,Info,Loader2),
	    autoloader_simplify((Loader1,Loader2),Loader)
	;   T = (A ## B) ->
	    tag_autoloader(Name,A,Info,Loader1),
	    tag_autoloader(Name,B,Info,Loader2),
	    autoloader_simplify((Loader1,Loader2),Loader)
	;   T= (A;B) ->
	    tag_autoloader(Name,A,Info,Loader1),
	    tag_autoloader(Name,B,Info,Loader2),
	    autoloader_simplify((Loader1 xor Loader2),Loader)
	;   T = tag_node{ label=>L::[_|_],kind=>scan } ->
	    Loader = phrase(L,_,_)
	;   T = tag_node{ kind=>anchor, label=>Label, top => Top, token => Token } ->
	    Info = tag_anchor{ family => Family },
	    Loader = tag_family_load(Family,Label,Top,Token,Name)
	;   T = tag_node{ kind=>coanchor, label=>Label, top => Top, token => Token } ->
	    Loader = tag_check_coanchor(Label,Top,Token)
	;   T = tag_node{ children => Children::[_|_] } ->
	    tag_autoloader(Name,Children,Info,Loader)
	;   T = guard{ goal => A, minus => fail } ->
	    tag_autoloader(Name,A,Info,Loader)
	;   
	    Loader = true
	)
	.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Interleave operator

%% We essentialy implement interleaving for TAG by copy-paste of interleaving for DCGs
%% It would best in some future to use some object-oriented mechanism
%% to get a more compact notation

%%! @case Interleave node
tag_compile_subtree_handler(Id,
			    Body::(A ## B),
			    Left,Right,Cont,Trans,Type,Info) :-
	interleave_initialize,
	il_label([],Label),
	tag_numbervars([IL_Left,IL_Right],Id,_),
	il_tupple(Body,Tupple),
	tag_il_loop_to_lpda(Id,Left,IL_Right,Loop_Code,Type,Info),
	tag_il_body_to_lpda(Id,Body,
			    IL_Left,IL_Right,IL_Out,IL_Out,
			    noop,Loop_Code,
			    Trans1,Type,Info,Label),
	tag_il_start_to_lpda(Id,
			     Cont,
			     Right,[],
			     [],IL_Left,
			     Trans1,
			     Trans2,
			     Type,Info,Label),
	Trans = ( '*NOOP*'(Tupple) :> '*IL-START*'(Trans2) ),
%%	format('IL Trans: ~K\n',[Trans]),
	true
	.

:-std_prolog il_tupple/2.

il_tupple(Body,Tupple) :-
	il_all_tupples(Body,[],Tupples),
	il_combine_tupples(Tupples,Tupple)
	.

:-std_prolog il_all_tupples/3.

il_all_tupples(Body,T1,T2) :-
	( Body = (A ## B) ->
	  il_all_tupples(A,T1,T3),
	  il_all_tupples(B,T3,T2)
	; tupple(Body,T),
	  T2 = [T|T1]
	)
	.

:-rec_prolog
	il_combine_tupples/2,
	il_combine_tupples_aux/3.

il_combine_tupples([],Null) :-
	null_tupple(Null).
%%il_combine_tupples([T],[]).
il_combine_tupples([T1|L],XT) :-
	il_combine_tupples(L,XL),
	il_combine_tupples_aux(T1,L,XT1),
	union(XT1,XL,XT)
	.

il_combine_tupples_aux(T,[],Null) :- null_tupple(Null).
il_combine_tupples_aux(T,[T1|L],XT) :-
	il_combine_tupples_aux(T,L,XL),
	inter(T,T1,T2),
	union(T2,XL,XT).

%%! @predicate tag_il_loop_to_lpda(Id,Pos,IL_Pos,Trans,Type,Info)
%%! @brief Generate LPDA code to select an interleave alternative
:-std_prolog tag_il_loop_to_lpda/6.

tag_il_loop_to_lpda(Id,Pos,IL_Pos,'*IL-LOOP-WRAP*'(Trans),Type,Info) :-
	tag_numbervars([IL_Loop,Loop_Jump],Id,_),
	tag_il_args(IL_Loop,Pos,IL_Args_Loop),
	update_counter(il_choice,Label),
	body_to_lpda(Id,
		     (	 %% format('Ready to choose ~w left=~w il=~w\n',[Label,Pos,IL_Pos]),
			 interleave_choose(Label,IL_Pos,IL_Loop,Loop_Jump)
			 %% format('Going to loop il_loop=~w\n',[IL_Loop])
		     ),
		     '*IL-LOOP*'(Loop_Jump,IL_Args_Loop),
		     Trans,
		     Type)
	.

%%! @predicate tag_il_start_to_lpda(Id,Exit,Pos,IL_Pos,IL_Left,IL_Right,Cont,Trans,Type,Info,Label)
%%! @brief Generate LPDA code to start a new interleave block with its exit alternative
:-std_prolog tag_il_start_to_lpda/11.

tag_il_start_to_lpda( Id,
		      Exit,
		      Pos,IL_Pos,
		      IL_Left,IL_Right,
		      IL_Cont,
		      '*IL-ALT*'(Exit_Jump,Exit,IL_Args_Exit,[]) :> Trans1,
		      Type,Info,Label
		    ) :-
	tag_numbervars(Exit_Jump,Id,_),
	tag_il_args(IL_Pos,Pos,IL_Args_Exit),	
	body_to_lpda(Id,
		     interleave_start(Label,Exit_Jump,IL_Left,IL_Right),
		     IL_Cont,
		     Trans1,
		     Type)
	.

%%! @predicate tag_il_register_to_lpda( Id,Alt,Pos,IL_Pos,IL_Left,IL_Right,Cont,Trans,Type,Info,Label) 
%%! @brief Generate LPDA code to register an interleave alternative
:-std_prolog tag_il_register_to_lpda/12.

tag_il_register_to_lpda( Id,
			 Alt,
			 Pos,IL_Pos,
			 IL_Left,IL_Right,
			 IL_Cont,
			 VLoop,
			 '*IL-ALT*'(Alt_Jump,Alt,IL_Args_Alt,VLoop) :> Trans1,
			 Type,Info,Label
		       ) :-
	my_numbervars(Alt_Jump,Id,_),
	tag_il_args(IL_Pos,Pos,IL_Args_Alt),	
	body_to_lpda(Id,
		     interleave_register(Label,Alt_Jump,IL_Left,IL_Right),
		     IL_Cont,
		     Trans1,
		     Type)
	.

%% inherited from dcg.pl
:-light_tabular interleave_initialize/0.


%% inherited from dcg.pl
:-std_prolog il_label/2.


%%! @predicate tag_il_args(IL_Table,Left,IL_Args)
%%! @brief Collect arguments for TAG interleaving alternative
:-light_tabular tag_il_args/3.

tag_il_args(IL_Table,Left,IL_Args::[IL_Table,Left]).

%%! @predicate tag_il_body_to_lpda(Id,A,IL_Left,IL_Right,IL_Out_Left,IL_Out_Right,Cont,Trans,Type,Info,Label)
%%! @brief essentially a wrapper around tag_il_body_to_lpda_handler/11
%%! Also wrap interleaving registration of an alternative build with tag_body_to_lpda/9

%%! @param Id  -- Id of TAG tree being compiled, used to number variables
%%! @param A -- The TAG subtree being compiled
%%! @param IL_Left -- The IL table before registering A
%%! @param IL_Right -- The IL table after registering A
%%! @param IL_Out_Left -- The IL table when starting alternative A
%%! @param IL_Out_Right -- The IL table before selecting next alternative Cont after A
%%! @param Cont -- Next alternative after A (possibly empty)
%%! @param Trans -- Resulting LPDA fragment
%%! @param Type -- Compilation mode
%%! @param Info -- Info of the tag tree 
%%! @param Label -- Label of current interleaving block 

:-std_prolog tag_il_body_to_lpda/12.

tag_il_body_to_lpda(Id,Body1,
		    IL_Left,IL_Right,IL_Out_Left,IL_Out_Right,
		    Cont, IL_Cont,
		    Trans,Type,Info,Label
		   ) :-
	(   toplevel(tag_term_expand(Body1,Body)),
	    my_numbervars(Body,Id,_)
	xor Body1=Body), % *HOOK*
	( tag_il_body_to_lpda_handler(Id,Body,
				      IL_Left,IL_Right,IL_Out_Left,IL_Out_Right,
				      Cont,IL_Cont,
				      Trans,Type,Info,Label ) ->
	    true
	;
	    tag_numbervars([Left,Right,IL_Out_Left,IL_Out_Right],Id,_),
	    tag_il_loop_to_lpda(Id,Right,IL_Out_Right,Loop_Code,Type,Info),
	    ( Cont == noop ->
		Cont1 = Loop_Code,
		Right=_Right,
		VLoop=[]
	    
	    ;	Cont = [VLoop^Loop_Code,_Right,Right]^KK ->
%%		format('IL Case ~w\n',[Cont]),
		Cont1 = KK
	    ;	
		Cont1 = Cont :> Loop_Code,
		Right=_Right,
		VLoop=[]
	    ),
	    tag_compile_subtree(Id,Body, Left, _Right,Cont1, Trans1,Type,Info),
	    tag_il_register_to_lpda( Id,Trans1,
				     Left,IL_Out_Left,
				     IL_Left,IL_Right,IL_Cont,
				     [Loop_Code|VLoop],
				     Trans,
				     Type,Info,Label
				   )
	)
	.

%%! @brief to express handlers for all cases of TAG constructions whithin a interleaving block

:-rec_prolog tag_il_body_to_lpda_handler/12.


%%! @case Empty Scan []
tag_il_body_to_lpda_handler(Id,
			    % [],
			    tag_node{ label => [], kind => scan },
			    IL_Left,IL_Right,_,_,
			    Cont,IL_Cont,
			    Trans:: (unify(IL_Left,IL_Right) :> IL_Cont),
			    Type,Info,Label
			   ) :-
%%	format('IL Void Scan ~w\n',[Trans]),
	true
	.


%%! @case Concatenation (A,B)
tag_il_body_to_lpda_handler(Id,(A,B),
			    IL_Left,IL_Right,IL_Out_Left,IL_Out_Right,
			    Cont,IL_Cont,
			    Trans,Type,Info,Label
			   ) :-
	tag_numbervars([IL_LeftB,IL_RightB],Id,_),
	tag_il_body_to_lpda(Id,B,
			    IL_LeftB,IL_RightB,IL_Out_Left,IL_Out_Right,
			    Cont,noop,
			    TransB,Type,Info,Label),
	tag_il_body_to_lpda(Id,A,
			    IL_Left,IL_Right,IL_LeftB,IL_RightB,
			    TransB,IL_Cont,
			    Trans,Type,Info,Label)
	.

%%! @case Disjunction (A;B)
tag_il_body_to_lpda_handler(Id,(A;B),
			    IL_Left,IL_Right,IL_Out_Left,IL_Out_Right,
			    Cont,IL_Cont,
			    Trans,Type,Info,Label
			   ) :-
	\+ A = (_ -> _),
	handle_choice(IL_Cont,Trans,IL_Last,ContA,ContB),
	tag_il_body_to_lpda(Id,A,
			    IL_Left,IL_Right,IL_Out_Left,IL_Out_Right,
			    Cont,IL_Last,
			    ContA,Type,Info,Label),
	tag_il_body_to_lpda(Id,B,
			    IL_Left,IL_Right,IL_Out_Left,IL_Out_Right,
			    Cont,IL_Last,
			    ContB,Type,Info,Label)
	.


%%! @case Guarded node
tag_il_body_to_lpda_handler(Id,
			    guard{ goal => A,
				   plus => Plus,
				   minus => Minus
				 },
			    IL_Left,IL_Right,IL_Out_Left,IL_Out_Right,
			    Cont,IL_Cont,
			    Trans,Type,Info,Label
			   ) :-
	\+ var(Minus),
	\+ (Minus == fail),
	handle_choice(IL_Cont,Trans,IL_Last,ContA,ContB),
	body_to_lpda(Id,Plus,IL_Last,IL_LastA,Type),
	tag_il_body_to_lpda(Id,
%%			    guard{ goal => A, plus => Plus, minus => fail },
			    A,
			    IL_Left,IL_Right,IL_Out_Left,IL_Out_Right,
			    Cont,IL_LastA,
			    ContA,Type,Info,Label),
	body_to_lpda(Id,Minus,unify(IL_Left,IL_Right) :> IL_Last,ContB,Type),
%%	format('IL Guard ~w\n',[Trans]),
	true
	.


%%! @case Interleaving (A ## B)
tag_il_body_to_lpda_handler(Id,(A ## B),
			    IL_Left,IL_Right,IL_Out_Left,IL_Out_Right,
			    Cont,
			    IL_Cont,
			    Trans,Type,Info,Label
			   ):-
	( Cont == noop ->  
	    Label2 = Label,
	    IL_Left2 = IL_Left,
	    IL_Right2=IL_Right,
	    Trans = TransA
	;   
	    il_label(Label,Label2),
	    tag_numbervars([Right,IL_Left2,IL_Right2],Id,_),
	    tag_il_loop_to_lpda(Id,Right,IL_Out_Right,Loop_Code,Type,Info),
	    tag_il_start_to_lpda(Id,
				 Cont :> Loop_Code,
				 Right,IL_Out_Left,
				 IL_Left,IL_Left2,
				 TransA,
				 Trans,
				 Type, Info, Label2)
	),
	tag_numbervars(IL_Middle,Id,_),
	tag_il_body_to_lpda(Id,B,
			    IL_Middle,IL_Right,IL_OutB,IL_OutB,
			    noop,IL_Cont,
			    TransB,Type,Info,Label2
			   ),
	tag_il_body_to_lpda(Id,A,
			    IL_Left2,IL_Middle,IL_OutA,IL_OutA,
			    noop,TransB,
			    TransA,Type,Info,Label2
			   )
	.

%% @case Kleene Star (A @*)
tag_il_body_to_lpda_handler(Id,(AA @*),
			    IL_Left,IL_Right,
			    IL_Out_Left,IL_Out_Right,
			    Cont,IL_Cont,
			    Trans,Type,Info,Label
			   ) :-
	(   AA = Vars ^ A xor AA = A),
	tag_il_body_to_lpda_handler(Id,@*{goal => A, vars=>Vars},
				    IL_Left,IL_Right,
				    IL_Out_Left,IL_Out_Right,
				    Cont,IL_Cont,
				    Trans,Type,Info,Label
				   )
	.

%% @case Range Kleene Star @*(A,[Min,Max],First^Arg^Param^Last)
tag_il_body_to_lpda_handler(Id,
			    K,
			    IL_Left,IL_Right,
			    IL_Out_Left,IL_Out_Right,
			    Cont,IL_Cont,
			    Trans,Type,Info,Label
			   ) :-
%%	format('IL KLEENE 2 ~w out_left=~w out_right=~w cont=~w\n',[AA,IL_Out_Left,IL_Out_Right,Cont]),
	kleene_normalize(K,
			 KleeneStruct:: @*{ goal => A,
					    vars => Vars,
					    from => From,
					    to => To,
					    collect_first=>UserFirst,
					    collect_loop=>UserParam,
					    collect_next=>UserArg,
					    collect_last=>UserLast
					  }),
	update_counter(kleene, KleeneId),
	name_builder('_kleene~w',[KleeneId],KleeneLabel),
	kleene_conditions(KleeneCond::[From,To],[Counter,CounterInc],InitCond,ExitCond,LoopCond),
	tag_numbervars([Counter,CounterInc,InitCond,ExitCond,LoopCond,KleeneStruct],Id,_),
	tag_numbervars([IL_Out_Right,IL_Out_Left,
			IL_Left2,IL_Right2,
			IL_Left3],Id,_),

	tag_numbervars(Right,Id,_),
	Args1 = [Counter,IL_Left2,_Right],
	Loop_Args1 = [CounterInc,IL_Left3,Right],
	append(Args1,UserParam,Args),
	append(Loop_Args1,UserArg,Loop_Args),
	tupple(Vars,TA),
	tupple(Cont,TCont),
	extend_tupple([IL_Out_Left,IL_Out_Right,IL_Right,UserLast,KleeneCond],TCont,RTCont),
	body_to_lpda(Id, (  %% format('exit il=~w counter=~w\n',[IL_Left2,Counter]),
			     ExitCond,
			     UserLast=UserParam,
			     IL_Left2=IL_Out_Left,
			     IL_Right2=IL_Out_Right			     
			 ),
		     Cont,
		     ContExit,
		     Type),

	VLoop=[IL_Left2,IL_Right2,IL_Left3,IL_Right,IL_Left,IL_Out_Right,IL_Out_Left],
	Kleene_Last= [VLoop^_Loop,
		      Right,
		      _Right]^ '*KLEENE-LAST*'(KleeneLabel,
					       Loop_Args,
					       Args,
					       TA,
					       RTCont,
					       _Loop
					      ),
	
	tag_il_body_to_lpda(Id,
			    A,
			    IL_Left2,IL_Right2,
			    IL_Left3,IL_Right2,
			    Kleene_Last,
			    noop,
			    ContLoop2,
			    Type,
			    Info,
			    Label),
	body_to_lpda(Id,
		     (
			 %% format('loop il=~w\n',[IL_Left2]),
			 LoopCond
		     ),
		     ContLoop2,
		     ContLoop,
		     Type),


	Trans1 = '*KLEENE*'(KleeneLabel,
			    TA,
			    '*ALTERNATIVE*'(ContExit,ContLoop)
			    ),

	body_to_lpda(Id,
		     IL_Right=IL_Right2,
		     IL_Cont,
		     IL_Cont2,
		     Type),
	
	body_to_lpda(Id,
		     (%%	 format('enter il=~w\n',[IL_Left]),
			 InitCond,
			 UserFirst=UserParam,
			 IL_Left=IL_Left2
			 ),
		     Trans1 :> IL_Cont2,Trans,Type),
	
%%	format('\nIL Kleene ~K\n',[Trans]),
	true
	.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Reader

directive(D::tag_custom_subst(Pred),_,_,_) :-
	directive_updater( tag_custom_subst(_),
			   tag_custom_subst(Pred) )
	.

