/* -*- mode:prolog; -*-
 ******************************************************************
 * $Id: tag_normalize.pl 724 2008-09-17 11:35:41Z clerger $
 * Copyright (C) 1998, 2002, 2003, 2004, 2005, 2008 by INRIA 
 * Authors:
 *      Djame Seddah <Djame.Seddah@inria.fr>
 *      Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>
 * ----------------------------------------------------------------
 *
 *  tag_normalize.pl -- TAG normalizer
 *
 * ----------------------------------------------------------------
 * Description
 *  Normalize Tree Adjoining Grammars
 * ----------------------------------------------------------------
e */

:-include 'header.pl'.

:-rec_prolog
	normalize/5,
	normalize_args/5,
	feature_normalize/3
	.

:-finite_set(adj,[no,yes,strict]).

normalize(Tree_with_Modifier,
	  Normalized_Tree,
	  Left,
	  Right,
	  Tree_Name
	 ):-
	Skel = tag_node{ id => Id,
			 label => Label,
			 children => Children,
			 kind => Kind,
			 adj => Adj::adj[],
			 spine => Spine,
			 top=>Top,
			 bot=>Bot
		       },
	( %% SubstNode with variable label
	    (	var(Tree_with_Modifier),  Label = '$call'(Tree_with_Modifier)
		xor	Tree_with_Modifier = '$VAR'(_),  Label = '$call'(Tree_with_Modifier)
		xor Tree_with_Modifier = '$call'(_), Label = Tree_with_Modifier
	    ) ->
	    Normalized_Tree = Skel,
	    (	Adj = no xor error( 'Not a valid TAG adjonction node ~w',[Tree_with_Modifier])),
	    Kind=subst,
	    Left=Right,
	    set_spine(Left,Spine),
	    misc_check(Normalized_Tree,Tree_with_Modifier,Tree_Name)
	%% ---------
	%% Meta Call
	;   Tree_with_Modifier =.. [apply,Label1|Rest] ->
	    Normalized_Tree = Skel,
	    Label = '$call'(Label1),
	    normalize_args(Rest,Children,Left,Right,Tree_Name),
	    ( Left \== Right -> Spine = yes ; set_spine(Left,Spine) ),
	    misc_check(Normalized_Tree,Tree_with_Modifier,Tree_Name)
	%% ------------
	%% Disjunction
	;   
	    Tree_with_Modifier = (T1;T2) ->
	    Normalized_Tree = (NT1;NT2),
	    (	normalize(T1,NT1,Left,Right,Tree_Name),
		normalize(T2,NT2,Left,Right,Tree_Name)
	    xor error('Incompatible TAG alternatives on ~w\n',[T])
	    )
	%% -----------
	%% guards
	;
	    Tree_with_Modifier = guard{ goal => T1, plus => Plus, minus => Minus } ->
	    (	Plus = true xor true),
	    (	Minus = fail xor true),
	    normalize(T1,NT1,Left,Right,Tree_Name),
	    Normalized_Tree = guard{ goal => NT1, plus => Plus, minus => Minus }
	%% ------------
	%% Optional
	;   Tree_with_Modifier = (T1 ?) ->
	  normalize(T1,NT1,Left,Right,Tree_Name),
	  Normalized_Tree = guard{ goal => NT1, plus => true, minus => true}
%%	    normalize(([];T1),Normalized_Tree,Left,Right,Tree_Name)
	%% ------------
	%% Sequence
	;   Tree_with_Modifier = (T1,T2) ->
	    Normalized_Tree = [NT1,NT2],
	    normalize(T1,NT1,Left,Middle,Tree_Name),
	    normalize(T2,NT2,Middle,Right,Tree_Name)
	%% ------------
	%% Kleene
	;   Tree_with_Modifier = (T1 @*) ->
	    normalize(@*{ goal => T1},Normalized_Tree,Left,Right,Tree_Name)
	%% ------------
	%% Generalized Kleene
	;   Tree_with_Modifier = @*{} ->
%%	    format('Normalize Kleene ~w\n',[Tree_with_Modifier]),
	    kleene_normalize(Tree_with_Modifier,
			     @*{ goal => T1,
				 vars => K_Vars,
				 from => K_From,
				 to => K_To,
				 collect_first => KCF,
				 collect_last => KCL,
				 collect_loop => KCLoop,
				 collect_next => KCN,
				 collect_pred => KCP
			       }),
	    Normalized_Tree = @*{ goal => NT1,
				  vars => K_Vars,
				  from => K_From,
				  to => K_To,
				  collect_first => KCF,
				  collect_last => KCL,
				  collect_loop => KCLoop,
				  collect_next => KCN,
				  collect_pred => KCP
				},
	    normalize(T1,NT1,Left,Right,Tree_Name),
%%	    format('\t-->~w\n',[Normalized_Tree]),
	    (	Left == Right
	    xor error('Kleene star not allowed around TAG spine nodes: ~w\n',
		      [T1])
	    )
	%% -----------
	%% Interleave
	; Tree_with_Modifier = (T1 ## T2) ->
	    normalize(T1,NT1,Left,Middle,Tree_Name),
	    normalize(T2,NT2,Middle,Right,Tree_Name),
	    Normalized_Tree = (NT1 ## NT2)
	%% ------------
	%% Obligatory adjonction node
	;   Tree_with_Modifier = ++ Tree ->
	    Normalized_Tree = Skel,
	    (	Adj = strict xor error( 'Multiple incompatible marks on ~w',[Tree_with_Modifier])),
	    normalize(Tree,Normalized_Tree,Left,Right,Tree_Name)
	%% ------------
	%% Non obligatory adjonction node
	;   Tree_with_Modifier = +Tree ->
	    Normalized_Tree = Skel,
	    (	Adj = adj[yes,strict] xor error( 'Conflict on TAG modifiers ~w',[Tree_with_Modifier])),
	    normalize(Tree,Normalized_Tree,Left,Right,Tree_Name)
	%% ------------
	%% Non adjonction node
	;   Tree_with_Modifier = -Tree ->
	    Normalized_Tree = Skel,
	    (	Adj = no xor error( 'Multiple incompatible marks on ~w',[Tree_with_Modifier])),
	    normalize(Tree,Normalized_Tree,Left,Right,Tree_Name)
	%% ------------
	%% Anchor node (at most one per tree :: don't how to ensure it !)
	;   Tree_with_Modifier = (<> Label1) ->
	    Normalized_Tree = Skel,
	    ( normalize_label(Label1,Label) xor error( 'Not a valid TAG anchor ~w', [Tree_with_Modifier])),
	    ( \+ recorded( tag_anchor(Tree_Name,_) ) xor error( 'More than one anchor in tree ~w: ~w', [Tree_Name,Tree_with_Modifier])),
	    Kind=anchor,
	    Left = Right,
	    set_spine(Left,Spine),
	    misc_check(Normalized_Tree,Tree_with_Modifier,Tree_Name),
	    record( tag_anchor(Tree_Name,Id) )
	%% ------------
	%% CoAnchor node (almost like anchor node and a lex node)
	;   Tree_with_Modifier = (<=> Label1) ->
	    Normalized_Tree = Skel,
	    ( normalize_label(Label1,Label) xor error( 'Not a valid TAG coanchor ~w', [Tree_with_Modifier])),
	    Kind=coanchor,
	    Left = Right,
	    set_spine(Left,Spine),
	    misc_check(Normalized_Tree,Tree_with_Modifier,Tree_Name)
	%% -------------
	%% Features
	;   Tree_with_Modifier = (Features at Tree) ->
	    Normalized_Tree = Skel,
	    feature_normalize(Features,Normalized_Tree,Tree),
	    normalize(Tree,Normalized_Tree,Left,Right,Tree_Name)
	%% -------------
	%% Foot node or SubstFoot node
	;   (Tree_with_Modifier = * Label1, Kind = foot
	    ;	Tree_with_Modifier = (= Label1), Kind = substfoot ) ->
	    Normalized_Tree = Skel,
	    (	normalize_label(Label1,Label) xor error( 'Not a valid TAG foot ~w', [Tree_with_Modifier])),
	    %% We do not allow adjonction on a foot node
	    (	Adj = no xor error( 'Not a valid TAG adjonction node ~w',[Tree_with_Modifier])),
	    %% Check that the foot node is the first foot node in an aux tree
	    %% ( we no longer check that the label equals the root label
	    %%   to allow for spinetrees. Actually, there is no need for a label
	    %%   on foot nodes ! )
	    ( Left == nospine ->
		error( 'foot ~w in elementary tree',[Tree_with_Modifier])
	    ;	Left == right ->
		error( 'more than one foot ~w: ~w',[Tree_Name,Tree_with_Modifier])
	    ;	Left = left(_) -> %% no check on label
		Right = right
	    ;	
		error( 'Not a valid TAG foot ~w ~w ~w',[Tree_with_Modifier,Left,Right])
	    ),
	    Spine=yes,
	    misc_check(Normalized_Tree,Tree_with_Modifier,Tree_Name)
	%% --------------
	%%  Scan node
	;   ( Tree_with_Modifier= []
	    ; Tree_with_Modifier = [_|_]
	    ) ->
	    Normalized_Tree = Skel,
	    (	Adj = no xor error( 'Not a valid TAG adjonction node ~w',[Tree_with_Modifier])),
	    Kind = scan,
	    Label = Tree_with_Modifier,
	    Children = [],
	    set_spine(Left,Spine),
	    Left = Right,
	    (Id=[] xor true),
	    (Top = [], Bot = [] xor error('Features on scan node',[Tree_with_Modifier]))
	%% --------------
	%% Substitution node
	;   atomic(Tree_with_Modifier) ->
	    %% No adj allowed on substitution mode
	    Normalized_Tree = Skel,
	    (	Adj = no xor error( 'Not a valid TAG adjonction node ~w',[Tree_with_Modifier])),
	    Kind=subst,
	    normalize_label(Tree_with_Modifier,Label),
	    Left = Right,
	    set_spine(Left,Spine),
	    misc_check(Normalized_Tree,Tree_with_Modifier,Tree_Name)
	%% --------------
	%% Escape node (escape to Prolog)
	;   Tree_with_Modifier = {Label} ->
	    Normalized_Tree = Skel,
	    (	Adj = no xor  error( 'Not a valid TAG adjonction node ~w',[Tree_with_Modifier])),
	    Kind = escape,
	    Children=[],
	    set_spine(Left,Spine),
	    Left=Right,
	    Id=[],
	    (	Top = [], Bot = [] xor error('Features on escape  node',[Tree_with_Modifier]))
	%% --------------
	%% Answer traversal
	;   Tree_with_Modifier = '$answers'(T1) ->
	    Normalized_Tree = '$answers'(NT1),
	    normalize(T1,NT1,Left,Right,Tree_Name)
	%% --------------
	%% Protected node
	;   Tree_with_Modifier = '$protect'(T1) ->
	    Normalized_Tree = '$protect'(NT1),
	    normalize(T1,NT1,Left,Right,Tree_Name)
	%% --------------
	%% Position Predicate
	; Tree_with_Modifier = '$pos'(_) ->
	    Normalized_Tree = Tree_with_Modifier,
	    Left = Right
	%% --------------
	%% Compound node
	;   Kind = std,
	    Normalized_Tree = Skel,
	    Tree_with_Modifier =.. [Label|Rest],
	    normalize_args(Rest,Children,Left,Right,Tree_Name),	    
	    ( Left \== Right -> Spine = yes ; set_spine(Left,Spine) ),
	    misc_check(Normalized_Tree,Tree_with_Modifier,Tree_Name)
	),
	abolish(tag_anchor/2),
	abolish(tag_node_id/2)
	.


:-std_prolog normalize_label/2.

normalize_label(Label1,Label) :-
	( var(Label1) -> Label = '$call'(Label)
	;   atomic(Label1),
	    Label=Label1
	)
	.

:-std_prolog set_spine/2.

set_spine(Left,Spine) :-
	( Left = nospine ->
	    Spine=no
	;   Left = right ->
	    Spine=right
	;   
	    Spine=left
	)
	.

normalize_args([],[],Left,Left,_).
normalize_args([Head|Args],[T|Q],Left,Right,Tree_Name):-
	normalize(Head,T,Left,Middle,Tree_Name),
	normalize_args(Args,Q,Middle,Right,Tree_Name).	    

feature_normalize(Features,Normalized_Tree::tag_node{ bot=> Bot, top=>Top, id=> Id, token => Token },Tree) :-
	( Features = (F1 and F2) ->
	    feature_normalize(F1, Normalized_Tree, Tree),
	    feature_normalize(F2, Normalized_Tree, Tree)
	;   Features = (top = F1) ->
	    ( var(F1) -> Top=F1
	    ;	
		functor(F1,Functor,N),
		term_current_module_shift(tag(Functor/N),_,F1,F),
		(   F = Top xor error('top feature conflict ~w with ~w at ~w',[F,Top,Tree]))
	    )
	;   Features = (bot = F1) ->
	    ( var(F1) -> Bot =F1
	    ;	functor(F1,Functor,N),
		term_current_module_shift(tag(Functor/N),_,F1,F),
		(   F = Bot xor error('bot feature conflict ~w with ~w at ~w',[F,Top,Tree]))
	    )
	;   Features = (id=F) ->
	    (	F = Id xor error('node id conflict ~w with ~w at ~w',[F,Id,Tree]))
	;   Features = (token=F) ->
	    (	F = Token xor error('node token conflict ~w with ~w at ~w',[F,Token,Tree]) )
	;   
	    error( 'not a valid feature ~w at ~w',[Features,Tree])
	)
	.

:-std_prolog feature_check/3.

feature_check(Label,Top,Bot) :-
	( Label = '$call'(_) ->
	    true
	;   recorded( tag_features(Label,T,B) ) ->
	    (	Top=T, Bot=B
	    xor error('features top ~w or bot ~w for ~w do not unify with defaults',[Top,Bot,Label]))
	;   
	    (	Label=Top,Label=Bot 
	    xor functor(Top,Label,N),functor(Bot,Label,N)
	    xor functor(Bot,Label,N),functor(Top,Label,N)
	    xor error('features top ~w or bot ~w not built on same non terminal ~w',[Top,Bot,Label])
	    )
	).

:-rec_prolog misc_check/3.

misc_check( Normalized_Tree :: tag_node{ id => Id,
					 label => Label,
					 children => Children,
					 adj => Adj,
					 spine => Spine,
					 top=>Top,
					 bot=>Bot,
					 kind => Kind,
					 token => Token
				       },
	    Tree_with_Modifier,
	    Tree_Name
	 ) :-
	( Adj==no ->
	    (	Top = Bot
	    xor error('Conflicting features top ~w and bot ~w on node ~w',[Top,Bot,Tree_with_Modifier]))
	;   Adj == strict ->
	    true
	;   \+ Top = Bot ->
	    Adj = strict
	;
	    Adj = yes
	),
	feature_check(Label,Top,Bot),
	name_tag_node(Id,Tree_Name),
	Spine ?= no,
	Children ?=[],
	(   domain(Kind,[anchor,coanchor])
	xor Token = []
	xor warning('Token info ~w useless in tree for node ~w',[Token,Tree_Name,Tree_with_Modifier])
	)
	.

:-rec_prolog name_tag_node/2.

%% Local name of node in its tree
%% Should be unique in current tree
name_tag_node( Id, Tree_Name ) :-
	( nonvar( Id ) ->
	    true
	;
	    update_counter( tag_nodeid(Tree_Name), Id )
	),
	( recorded( tag_node_id(Tree_Name,Id) ) ->
	    error( 'duplicate local name ~w for tag node in tree ~w', [Id,Tree_Name] )
	;   
	    record( tag_node_id(Tree_Name,Id) )
	)
	.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Wrapping function to normalization

:-light_tabular normalize/2.
:-mode(normalize/2,+(+,-)).

normalize( tag_tree{ name=> Name,
		     family=>Family,
		     tree => T },
	   L ) :-
	%% See file note2.txt for naming conventions
	( var(Family) ->
	    (	recorded( tag_family( Family ) ) ->
		true
	    ;	var(Name) ->
		recorded( main_file(_,info(_,Family,_,_)) )
	    ;	Family=Name
	    )
	;   true
	),
	( var(Name) ->
	    update_counter( tag_trees, Id ),
	    number_chars( Id, Char_Id ),
	    atom_chars( LL, [0't,0'r,0'e,0'e|Char_Id] ),
	    default_module(LL,Name)
	;   true
	),
	( recorded( tag_tree(Name,_) ) ->
	    error( 'two TAG trees with same name ~w\n',[Name])
	;   record( tag_tree(Name,Family) )
	),
	normalize( T, L, Name ),
	%% Cleaning
	erase( tag_node_id(Name,_) )
	.

:-rec_prolog normalize/3.

normalize((tree T),L,Name):-
	normalize(T,L,nospine,nospine,Name)
%%	,format('Normalized: ~w\n\t~w\n\n',[T,L])
	.

normalize((auxtree T),L,Name):-
	normalize(T,L::tag_node{ label=>Root, spine=>Spine },left(Root),Right,Name),
	(Spine == yes, Right == right xor error('missing foot node in ~w\n',[T]))
%%	,format('Normalized: ~w\n\t~w\n\n',[T,L])
	.

normalize((spinetree T),spinetree(L),Name):-
	normalize(T,L::tag_node{ label=>Root, spine=>Spine },left(Root),Right,Name),
	(Spine == yes, Right == right xor error('missing foot node in ~w\n',[T]))
	.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Modules

term_module_shift(Module,Pred1::tag(F1/N),tag(F/N),T1,T) :-
       	( recorded( module_import(Pred1,Module1) ) ->
	    Module1 \== []
	;   Module \== [],
	    atom_module(F1,[],_),
	    Module1=Module
	),
        atom_module(F,Module1,F1),
        length(Args,N),
        T1 =.. [F1|Args],
        T =.. [F|Args]
        .




